; Syntax: NASM, tab width = 8 spaces (or higher)
; - 8bitbubsy, 2023
;
; PIT_CLK      = 1193181.81818181 (recurring)
; NUM_WRITES   = 11 (gives perfectly round dividend number)
; BYTES_PER_KB = 1024
; SCREEN_W     = 320
; SCREEN_H     = 200

	CPU 8086
	ORG 100h

SCREEN_W	EQU 320
SCREEN_H	EQU 200

Main:
	call DetectVGA
	jnc .L0
	lea dx,ErrorText
	call PutString
	jmp ExitToDOS
.L0:	; ---------------------
	mov ax,0F00h
	int 10h
	mov [OldVideoMode],al
	mov ax,13h
	int 10h
	; ---------------------
	call BenchmarkVGA
	; ---------------------
	xor ah,ah
	mov al,[OldVideoMode]
	int 10h
	; ---------------------
	lea dx,ThruputText
	call PutString
	; ---------------------
	mov cx,[TicksHi]
	mov bx,[TicksLo]
	; ---------------------
	test cx,cx	; Low-clamp result to 1
	jnz .L1
	test bx,bx
	jnz .L1
	mov bx,1
.L1:	; ---------------------
	; AX:DX = (820312500 + (Ticks/2)) / Ticks
	; 820312500.0 = (PIT_CLK * SCREEN_W * SCREEN_H * NUM_WRITES) / BYTES_PER_KB
	; (perfectly round number)
	; ---------------------
	mov dx,cx
	mov ax,bx
	shr dx,1
	rcr ax,1
	add ax,0F9B4h	; lo-word of 820312500
	adc dx,030E4h	; hi-word of 820312500
	call DivU32
	test dx,dx
	jnz .mbps
	cmp ax,1000
	jae .mbps
	; ---------------------
.kbps:	call PrintKbps
	jmp ExitToDOS
.mbps:	; ---------------------
	; AX:DX /= 1024 (result clamped to 65535 and put in DI)
	push ax
	mov cx,10
.loop:	shr dx,1
	rcr ax,1
	loop .loop
	test dx,dx
	jz .L2
	mov ax,65535
.L2:	mov di,ax
	pop ax
	; ---------------------
	; Get fractional part of AX:DX / 1024, and scale to 0..999
	; ---------------------
	mov cx,1024
	div cx
	mov ax,dx
	mov bx,1000
	mul bx
	mov bx,1024
	div bx
	; ---------------------
	mov bx,ax
	mov ax,di
	call PrintMbpsWithFrac
ExitToDOS:
	mov ah,4Ch
	int 21h

PutString:
	mov ah,9
	int 21h
	ret

HandleDigit:
	add al,'0'
	mov [di],al
	inc di
	xchg ax,dx
	xor dx,dx
	ret

PrintKbps:
	lea di,Dec3Str
	mov bx,100
	div bx
	call HandleDigit
	mov bx,10
	div bx
	call HandleDigit
	call HandleDigit
	; ---------------------
	lea di,Dec3Str
	mov cx,3
.loop:	cmp byte [di],'0'
	jne .done
	inc di
	jmp .loop
.done:	mov dx,di
	call PutString
	; ---------------------
	lea dx,EndTextKbps
	call PutString
	ret

	; AX = high part
	; BX = low part
PrintMbpsWithFrac:
	push bx
	xor dx,dx
	lea di,Dec5Str
	mov bx,10000
	div bx
	call HandleDigit
	mov bx,1000
	div bx
	call HandleDigit
	mov bx,100
	div bx
	call HandleDigit
	mov bx,10
	div bx
	call HandleDigit
	call HandleDigit
	; ---------------------
	lea di,Dec5Str
	mov cx,5
.loop:	cmp byte [di],'0'
	jne .done
	inc di
	jmp .loop
.done:	mov dx,di
	call PutString
	; ---------------------
	; Handle fractional part
	; ---------------------
	lea dx,DotText
	call PutString
	; ---------------------
	pop ax
	xor dx,dx
	lea di,Dec3Str
	mov bx,100
	div bx
	call HandleDigit
	mov bx,10
	div bx
	call HandleDigit
	call HandleDigit
	; ---------------------
	lea dx,Dec3Str
	call PutString
	; ---------------------
	lea dx,EndTextMbps
	call PutString
	ret

DetectVGA:
	mov ax,1A00h
	int 10h
	cmp al,26
	jne .error
	cmp bl,1
	je .ok
	cmp bl,7
	je .ok
	cmp bl,8
	je .ok
	cmp bl,10
	je .ok
	cmp bl,11
	je .ok
	cmp bl,12
	je .ok
.error:	stc
	ret
.ok:	clc
	ret

	; Disassembled 32-bit division routine from Open Watcom V2's runtime.
	; github.com/open-watcom/open-watcom-v2/tree/master/bld/clib/cgsupp/a/i4d.asm
	;
	; Input:
	;  DX:AX = Dividend
	;  CX:BX = Divisor
	;
	; Output:
	;  DX:AX = Quotient
	;  CX:BX = Remainder
	;
DivU32:
	test cx,cx
	jnz .L2
	dec bx
	jz .L1
	inc bx
	cmp bx,dx
	ja .L0
	mov cx,ax
	mov ax,dx
	xor dx,dx
	div bx
	xchg ax,cx
.L0:	div bx
	mov bx,dx
	mov dx,cx
	xor cx,cx
.L1:	ret
.L2:	cmp cx,dx
	jc .L4
	jnz .L3
	cmp bx,ax
	ja .L3
	sub ax,bx
	mov bx,ax
	xor cx,cx
	xor dx,dx
	mov ax,1
	ret
.L3:	xor cx,cx
	xor bx,bx
	xchg ax,bx
	xchg dx,cx
	ret
.L4:	push bp
	push si
	xor si,si
	mov bp,si
.L5:	add bx,bx
	adc cx,cx
	jc .L8
	inc bp
	cmp cx,dx
	jc .L5
	ja .L6
	cmp bx,ax
	jbe .L5
.L6:	clc
.L7:	adc si,si
	dec bp
	js .LB
.L8:	rcr cx,1
	rcr bx,1
	sub ax,bx
	sbb dx,cx
	cmc
	jc .L7
.L9:	add si,si
	dec bp
	js .LA
	shr cx,1
	rcr bx,1
	add ax,bx
	adc dx,cx
	jnc .L9
	jmp .L7
.LA:	add ax,bx
	adc dx,cx
.LB:	mov bx,ax
	mov cx,dx
	mov ax,si
	xor dx,dx
	pop si
	pop bp
	ret

PITOverflowInterrupt:
	inc bp
	push ax
	mov al,20h
	out 20h,al
	pop ax
	iret

%macro DO_FRAMEWRITE 0
	mov di,dx
	mov cx,bx
	rep stosw
%endmacro

BenchmarkVGA:
	push ax
	push bx
	push cx
	push dx
	push si
	push di
	push bp
	push es
	; ---------------------
	cli
	; ---------------------
	in al,21h		; disable IRQ 0
	mov [OldIRQ0Mask],al
	or al,1
	out 21h,al
	; ---------------------
	xor ax,ax		; set ES to segment zero
	mov es,ax
	mov ax,es:[32+0]	; set new interrupt 8 vector
	mov [OldInt8Off],ax
	mov ax,es:[32+2]
	mov [OldInt8Seg],ax
	mov ax,PITOverflowInterrupt
	mov es:[32+0],ax 
	mov es:[32+2],cs
	; ---------------------
	in al,21h		; enable IRQ 0
	and al,0FEh
	out 21h,al
	; ---------------------
	mov ax,0A000h
	mov es,ax
	mov bx,(SCREEN_W*SCREEN_H)/2
	cld
	; ---------------------	
	mov al,34h		; set PIT channel 0 (mode 2, rate generator)
	out 43h,al
	xor ax,ax		; clear upper byte too, for rep stosw
	out 40h,al
	out 40h,al
	; ---------------------
	sti
	; ---------------------
	xor dx,dx
	xor bp,bp		; clear bp (TicksHi) here, fix premature interrupt
	; ---------------------
	; do 22 iterations of frame writing (for good measure)
	; ---------------------
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	DO_FRAMEWRITE
	; ---------------------
	mov [TicksHi],bp
	cli
	; ---------------------
	mov al,34h		; read PIT channel 0 (mode 2, rate generator)
	out 43h,al
	in al,40h
	mov ah,al
	in al,40h
	xchg al,ah
	neg ax			; invert count (count down -> count up)
	mov [TicksLo],ax
	; ---------------------
	xor ax,ax		; set ES to segment zero
	mov es,ax
	mov ax,[OldInt8Off]	; set back old interrupt 8 vector
	mov es:[32+0],ax 
	mov ax,[OldInt8Seg]
	mov es:[32+2],ax
	; ---------------------
	mov al,[OldIRQ0Mask]	; set back old IRQ 0 mask
	out 21h,al
	; ---------------------
	mov al,36h		; reset PIT channel 0 (mode 3, count 0)
	out 43h,al
	xor al,al
	out 40h,al
	out 40h,al
	; ---------------------
	sti
	; ---------------------
	pop es
	pop bp
	pop di
	pop si
	pop dx
	pop cx
	pop bx
	pop ax
	retn

	ALIGN 2
OldInt8Off:	dw 0
OldInt8Seg:	dw 0
TicksLo:	dw 0
TicksHi:	dw 0
OldIRQ0Mask:	db 0
OldVideoMode:	db 0
Dec3Str:	db "000",'$'
Dec5Str:	db "00000",'$'
DotText:	db ".",'$'
ErrorText:	db "ERROR: This program requires a VGA graphics adapter.",13,10,'$'
ThruputText:	db "VGA thruput (16-bit writes): ",'$'
EndTextKbps:	db "kB/s",13,10,'$'
EndTextMbps:	db "MB/s",13,10,'$'
